using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Reflection;
using System.Collections;
using System.Threading;
using OtterUnitTestFramework.DataStruct;
using OtterUnitTestFramework.Public.Test;
using OtterUnitTestFramework.DataStruct.Reflection;
using OtterUnitTestFramework.Exceptions;
using OtterUnitTestFramework.Public;

namespace OtterUnitTestFramework.Forms
{
    /// <summary>
    /// This class implements the Form where the test is run and
    /// displayed.
    /// </summary>
    partial class OtterUnitTestFrameworkForm : Form
    {
        #region Constructors

        /// <summary>
        /// Default Constructor
        /// </summary>
        public OtterUnitTestFrameworkForm()
        {
            const int SPLASH_DISPLAY_TIME = 3000;
            //create splash and show it
            Thread splashThread = new Thread(new ThreadStart(ShowSplash));
            splashThread.Start();
            //initialize form
            InitializeComponent();
            //wait - show splash only
            Thread.Sleep(SPLASH_DISPLAY_TIME);
            //kill splash
            splashThread.Abort();
        }

        #endregion

        #region Button Listeners
        /// <summary>
        /// Select all Tests to be run.
        /// </summary>
        /// <param name="sender">Sender Object</param>
        /// <param name="e">Event Object.</param>
        private void BtnSelectAll_Click(object sender, EventArgs e)
        {
            //select all
            CheckOrUncheckAllMethodDisplays(true);
        }

        /// <summary>
        /// UnSelect all Tests to be run.
        /// </summary>
        /// <param name="sender">Sender Object</param>
        /// <param name="e">Event Object.</param>
        private void BtnUnselectAll_Click(object sender, EventArgs e)
        {
            //unselect all
            CheckOrUncheckAllMethodDisplays(false);
        }

        /// <summary>
        /// Execute Tests stored within Data Grid View Data Source. 
        /// </summary>
        /// <param name="sender">Sender Object</param>
        /// <param name="e">Event</param>
        private void BtnInitalizeTest_Click(object sender, EventArgs e)
        {
            //set cursor
            this.Cursor = Cursors.WaitCursor;
            //variables
            MethodInfo method = null;
            TestCaseTemplate testCaseTemplate = null;
            String message = String.Empty;
            Int32 displayedTests = 0; Int32 executedTests = 0;
            Int32 successfulTests = 0; Int32 failedTests = 0;
            bool isExpectingException = false;
            ExpectedExceptionAttribute expectedExceptionAttribute = null;
            //clean text bos
            _rtbTests.Text = String.Empty;
            //get data source
            IList<TestCaseTemplateDisplayer> dataSource = (IList<TestCaseTemplateDisplayer>)_dgvTests.DataSource;
            //get the number of displayed tests
            displayedTests = dataSource.Count;
            //go through the data source list and see which tests are to be executed
            if (dataSource != null && dataSource.Count > 0)
            {
                for (int index=0;index < dataSource.Count;index++)
                {
                    TestCaseTemplateDisplayer methodDiplayer = dataSource[index];
                    //check whether the test is to be executed
                    if (methodDiplayer.ExecuteTest)
                    {
                        //assign elements to execute the test
                        testCaseTemplate = methodDiplayer.TestCaseTemplate;
                        method = methodDiplayer.Method;
                        try
                        {
                            /* execute method */

                            //increment the counter of executed tests
                            executedTests++;
                            //get the method custom atributes
                            Object[] customAttributes = method.GetCustomAttributes(false);
                            //check whether it is being expected an exception
                            if (customAttributes != null && customAttributes.Length > 0)
                            {
                                foreach (Object attribute in customAttributes)
                                {
                                    if (attribute is ExpectedExceptionAttribute)
                                    {
                                        isExpectingException = true;
                                        expectedExceptionAttribute = (ExpectedExceptionAttribute) attribute;
                                        break;
                                    }
                                }
                            }

                            //in case an exception is expected treat it, otherwise, execute the method normally
                            if (isExpectingException)
                            {
                                ExecutetMethodExpectingException(testCaseTemplate, method, expectedExceptionAttribute.ExceptionTypes);
                            }
                            else
                            {
                                //execute test
                                ExecuteTestCaseTemplateMethod(method, testCaseTemplate);
                            }
                            //create message
                            message = "\n\n Test test of the method " + method.Name + " has passed!";
                            //log success message
                            _rtbTests.Text += message;
                            //update display element stating the test passed
                            methodDiplayer.HasPassed = TestApprovalStatus.HasPassed;
                            //update error message in the method displayer
                            methodDiplayer.ErrorMessage = message;
                            //increment the counter of successful tests
                            successfulTests++;
                        }
                        catch (AssertionException ae)
                        {
                            //create message
                            message = "\n\n The test of the method " + method.Name + " has failed. The error is: \nPlace where it has occured: " + ae.GetBaseException().Source + "\n Error Message: " + ae.GetBaseException().Message + "\n Stack Trace: " + ae.GetBaseException().StackTrace;
                            HandleExceptionInTestsExcecution(message, methodDiplayer);
                        }
                        catch (Exception ex)
                        {
                            //create message
                            message = "\n\n The test of the method " + method.Name + " has failed. The error is: \nPlace where it has occured: " + ex.GetBaseException().Source + "\n Error Message: " + ex.GetBaseException().Message + "\n Stack Trace: " + ex.GetBaseException().StackTrace;
                            HandleExceptionInTestsExcecution(message, methodDiplayer);
                        }
                        //reset test expectations
                        isExpectingException = false;
                        //update data source list
                        dataSource.RemoveAt(index);
                        dataSource.Insert(index, methodDiplayer);
                        //update data grid view data source
                        _dgvTests.DataSource = dataSource;
                        //refresh grid
                        _dgvTests.Refresh();
                    }
                }
                //calculate failed tests
                failedTests = executedTests - successfulTests;
                //show success message
                MessageBox.Show(this, "Tests executed successfully!\n\n Displayed Tests: " + displayedTests + "\n Executed Tests: " + executedTests + "\n Successful Tests: " + successfulTests + "\n Failed Tests: " + failedTests, "Otter Unit Test Framework", MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            else
            {
                String noTestErrorMessage = "No test was executed because none was registered in the selected Test class(es)!";
                //reset cursor
                this.Cursor = Cursors.Default;
                //put message in the Text Box
                _rtbTests.Text = noTestErrorMessage;
                //show message box
                MessageBox.Show(this, noTestErrorMessage, "Otter Unit Test Framework", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            }
            //reset cursor
            this.Cursor = Cursors.Default;
        }

        /// <summary>
        /// Close this Form.
        /// </summary>
        /// <param name="sender">Sender Object</param>
        /// <param name="e">Event</param>
        private void BtnClose_Click(object sender, EventArgs e)
        {
            //close this Form
            this.Close();
        }

        #endregion

        #region Form Listeners

        /// <summary>
        /// Event executed when the form is being closed. The closing operation
        /// can be cancelled here.
        /// </summary>
        /// <param name="sender">Sender Element</param>
        /// <param name="e">Closing Event</param>
        private void OtterUnitTestFrameworkForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            //make sure the user wants to quite
            DialogResult result = MessageBox.Show(this, "Are you sure you want to quit the application?", "Otter Unit Test Framework", MessageBoxButtons.YesNo, MessageBoxIcon.Question);
            //take action according to the user decision
            if (result == DialogResult.No)
            {
                //in case the answer is no, cancel the closing operation
                e.Cancel = true;
            }
            else
            {
                //state the appliation is being closed by setting the appropriate flag
                OtterUnitTestFrameworkApplication.IsApplicationBeingExecuted = false;
                //clean Test Case Template List to be tested
                TestCaseTemplateList.getInstance().Clear();
            }
        }

        /// <summary>
        /// Method executed before this form is loaded.
        /// </summary>
        /// <param name="sender">Sender Object</param>
        /// <param name="e">Event Object</param>
        private void OtterUnitTestFrameworkForm_Load(object sender, EventArgs e)
        {
            //set cursor
            this.Cursor = Cursors.WaitCursor;
            //get the list of classes to be added to the filter
            IList<String> classNameList = CreateClassNameList();
            /*
             * Add class names to the list. An important observation
             * is that the first added element is an empty String. This
             * is done because one may choose not to use the filter and
             * this empty String gives this option.
             */
            _cbxClassfilter.Items.Add(String.Empty);
            if (classNameList != null && classNameList.Count > 0)
            {
                foreach (String className in classNameList)
                {
                    //add element to the list
                    _cbxClassfilter.Items.Add(className);
                }
            }

            //create data grid view header
            CreatDataGridViewHeader();
            //create data grid view datasource with methods to be tested
            IList<TestCaseTemplateDisplayer> dataSource = CreateDataGridViewDataSourceUsingFilter(String.Empty);
            //populate data grid view
            _dgvTests.DataSource = dataSource;
            //reset cursor
            this.Cursor = Cursors.Default;
        }

        #endregion

        #region Grid View Listeners

        /// <summary>
        /// Show the message regarding a Test for a certain row.
        /// </summary>
        /// <param name="sender">Sender Object</param>
        /// <param name="e">Event Object</param>
        private void DgvTests_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            //only proceed in case there were a double-clicked element
            if (_dgvTests.CurrentRow != null)
            {
                //get clicked row
                DataGridViewRow clickedRow = _dgvTests.CurrentRow;
                //get the Test Case Template Displayer for this row
                TestCaseTemplateDisplayer templateDisplayer = (TestCaseTemplateDisplayer)clickedRow.DataBoundItem;
                //get the elements to be displayed
                String className = templateDisplayer.ClassName != null ? templateDisplayer.ClassName : String.Empty;
                String methodName = templateDisplayer.Method != null ? templateDisplayer.Method.Name : String.Empty;
                String message = templateDisplayer.ErrorMessage != null ? templateDisplayer.ErrorMessage : String.Empty;
                //show the message
                TestMessageForm messageForm = new TestMessageForm(className, methodName, message);
                messageForm.ShowDialog();
            }
            else
            {
                MessageBox.Show(this, "Please, select a Test to have its results displayed!", "Otter Unit Test Framework", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        #endregion

        #region Combo Box Listeners

        /// <summary>
        /// Execute the filter operation - filter the tests to be displayed
        /// by the selected Class.
        /// </summary>
        /// <param name="sender">Sender Object</param>
        /// <param name="e">Event</param>
        private void CbxClassfilter_SelectedIndexChanged(object sender, EventArgs e)
        {
            //set cursor for waiting
            this.Cursor = Cursors.WaitCursor;
            //only proceed in case there there were a selected element
            if (_cbxClassfilter.SelectedIndex >= 0)
            {
                //clean Test Log Rich Text Box
                _rtbTests.Text = String.Empty;
                //get selected class name - in case it is the empty string value, all classes will be selected.
                String className = (String)_cbxClassfilter.SelectedItem;
                //create data grid view datasource with methods to be tested, belonging to the filtered class
                IList<TestCaseTemplateDisplayer> dataSource = CreateDataGridViewDataSourceUsingFilter(className);
                //populate data grid view
                _dgvTests.DataSource = dataSource;
            }
            else
            {
                MessageBox.Show(this, "Please, select an element of the filter!", "Otter Unit Test Framework", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            //reset cursor
            this.Cursor = Cursors.Default;
        }

        #endregion

        #region Private Methods

        /// <summary>
        /// Display splash screen.
        /// </summary>
        private void ShowSplash()
        {
            SplashForm splashForm = new SplashForm();
            splashForm.ShowDialog();
        }

        /// <summary>
        /// Execute the Test Case method of a Test Case Template.
        /// </summary>
        /// <param name="method">Method to be executed</param>
        /// <param name="testCaseTemplate">Test Case Template Object</param>
        private static void ExecuteTestCaseTemplateMethod(MethodInfo method, TestCaseTemplate testCaseTemplate)
        {
            //call the start up
            testCaseTemplate.StartUp();
            //call the testCase method
            method.Invoke(testCaseTemplate, new Object[] { });
            //call the tear down method
            testCaseTemplate.TearDown();
        }

        /// <summary>
        /// Execute a test (from a Test Case Template class) using the method 
        /// entered as a parameter and expects an exception to be thrown.
        /// The exception must of one of the types entered as expectedExceptionTypes. 
        /// In case an exception is neither thrown nor is the type of one of the expectedExceptionTypes, 
        /// AssertException is dispatched.
        /// </summary>
        /// <param name="testCaseTemplate">Test Case Template holding the method to be executed</param>
        /// <param name="method">Method to be executed (IT MUST HAVE NO PARAMETERS)</param>
        /// <param name="expectedExceptionTypes">Exception types to be expected.</param>
        /// <exception cref="OtterUnitTestFramework.Exceptions.AssertionException">An exception is thrown in case the method execution does not produce one of the expected exceptions.</exception>
        private void ExecutetMethodExpectingException(TestCaseTemplate testCaseTemplate, MethodInfo method, Type[] expectedExceptionTypes)
        {
            bool hasThrownException = true;
            bool hasFoundAnExceptionMatch = false;
            Type baseExceptionType = null;
            try
            {
                //execute test
                ExecuteTestCaseTemplateMethod(method, testCaseTemplate);
                //an exception should have been thrown, in case it was not, mark it
                hasThrownException = false;
            }
            catch (Exception ex)
            {
                baseExceptionType = ex.GetBaseException() != null ? ex.GetBaseException().GetType() : null; 
                //the exception was thrown; check whether the correct exception was dispatched
                if (baseExceptionType != null)
                {
                    //go through the types of exceptions
                    if (expectedExceptionTypes != null)
                    {
                        IEnumerator expectedExceptionTypesEnum = expectedExceptionTypes.GetEnumerator();
                        while (expectedExceptionTypesEnum.MoveNext() && !hasFoundAnExceptionMatch)
                        {
                            //get expected exception
                            Type type = (Type) expectedExceptionTypesEnum.Current;
                            //get all the inherited types from the base exception
                            IList<Type> inheritedBaseExceptionTypeList = RetrieveInheritedSubTypes(baseExceptionType);
                            if (inheritedBaseExceptionTypeList != null && inheritedBaseExceptionTypeList.Count > 0)
                            {
                                foreach (Type inheritedBaseExceptionType in inheritedBaseExceptionTypeList)
                                {
                                    //in case the exceptions are of the same type, marke it and leave the loop
                                    if (type.Equals(inheritedBaseExceptionType))
                                    {
                                        hasFoundAnExceptionMatch = true;
                                        break;
                                    }
                                }
                            }
                        }
                    }
                    //in case no match was found, throw an exception
                    if (!hasFoundAnExceptionMatch)
                    {
                        throw new AssertionException("The Test has failed. The thrown exception was not of the one entered elements in the ExpectedException tag.");
                    }
                }
            }
            //in case no exception was thrown, throw one because this should not has happend
            if (!hasThrownException)
            {
                throw new AssertionException("The Test has failed. There was no thrown exception by the method Assert.DispatchException(Object, MethodInfo, Exception)");
            }
        }

        /// <summary>
        /// Retrieve the type and all the inherited types which the main
        /// one.
        /// </summary>
        /// <param name="type">Type of which sub types are retrieved</param>
        /// <returns>List of the type and its subtypes</returns>
        private IList<Type> RetrieveInheritedSubTypes(Type type)
        {
            IList<Type> typeList = null;
            //in case there is a base type, get it recursively
            if (type != null && type.BaseType != null)
            {
                //call recursion
                typeList = RetrieveInheritedSubTypes(type.BaseType);
            }
            /* after the recursion concatenate the types and sub types */

            //create the list in case it is null
            if (typeList == null)
            {
                typeList = new List<Type>();
            }
            //add the current type to the list
            if (type != null)
            {
                typeList.Add(type);
            }
            //return list of types
            return typeList;
        }

        /// <summary>
        /// Handle the exceptions blocks when one occurs in the method
        /// where the tests are performed.
        /// </summary>
        /// <param name="message">Error Message</param>
        /// <param name="methodDiplayer">Test Case Template Displayer element</param>
        private void HandleExceptionInTestsExcecution(String message, TestCaseTemplateDisplayer methodDiplayer)
        {
            //it means the test has failed, set the message to be printed
            _rtbTests.Text += message;
            //update display element stating the test NOT passed
            methodDiplayer.HasPassed = TestApprovalStatus.HasFailed;
            //update error message in the method displayer
            methodDiplayer.ErrorMessage = message;
        }


        /// <summary>
        /// Create the data grid view header.
        /// </summary>
        private void CreatDataGridViewHeader()
        {
            const int COLUMN_COUNT = 3;
            //populate data grid view header
            _dgvTests.ColumnCount = COLUMN_COUNT;

            DataGridViewCheckBoxColumn executeTestColumn = new DataGridViewCheckBoxColumn();
            _dgvTests.Columns.Insert(0, executeTestColumn);
            _dgvTests.Columns[0].Name = TestCaseTemplateHeaderDisplayer.EXECUTE_TEST;
            _dgvTests.Columns[1].Name = TestCaseTemplateHeaderDisplayer.CLASS_NAME;
            _dgvTests.Columns[2].Name = TestCaseTemplateHeaderDisplayer.METHOD_NAME;
            _dgvTests.Columns[3].Name = TestCaseTemplateHeaderDisplayer.HAS_PASSED;

            //make everything readonly except for the first column
            _dgvTests.Columns[1].ReadOnly = true;
            _dgvTests.Columns[2].ReadOnly = true;
            _dgvTests.Columns[3].ReadOnly = true;

            //relate columns of the data source to the object used as it
            _dgvTests.Columns[0].DataPropertyName = "ExecuteTest";
            _dgvTests.Columns[1].DataPropertyName = "ClassName";
            _dgvTests.Columns[2].DataPropertyName = "MethodName";
            _dgvTests.Columns[3].DataPropertyName = "HasPassedToBePrinted";
        }

        /// <summary>
        /// Create a List of Class Names based on the registered classes
        /// to have their methods tested.
        /// </summary>
        /// <returns>List of Class Names</returns>
        private IList<String> CreateClassNameList()
        {
            //list of class names to be returned
            IList<String> classNameList = null;
            //get Test Case Template List
            IList<TestCaseTemplate> testCaseTemplateList = TestCaseTemplateList.getInstance();
            //only proceed in case the list of tests are neither null nor empty
            if (testCaseTemplateList != null && testCaseTemplateList.Count > 0)
            {
                classNameList = new List<String>();
                foreach (TestCaseTemplate testCaseTemplate in testCaseTemplateList)
                {
                    classNameList.Add(testCaseTemplate.GetType().Name);
                }
            }
            //return list of class names
            return classNameList;
        }

        /// <summary>
        /// Create a Data Grid View Data Source with methods to be tested in the form of a list. Note
        /// that only methods that start with OtterUnitTestFramework.Test.TEST_METHOD_INITIAL_NAME
        /// will be used. Moreover, the method must have neither parameters nor returned value
        /// <br/> 
        /// Also note that a method parameter String classNameFilter is used as a filter
        /// of classes to be displayed. It means, that if a class name is entered as a parameter,
        /// the displayed methods will be the ones which belong to the filtered class. In case null
        /// or an empty string is entered, this filter will be ignored and all methods from all
        /// classes will be displayed.
        /// </summary>
        /// <param name="classNameFilter">In case it is neither null nor empty, will serve as the filter to display methods of only this class.</param>
        /// <returns>List holding objects to be used as a detaset for the data grid view</returns>
        private IList<TestCaseTemplateDisplayer> CreateDataGridViewDataSourceUsingFilter(String classNameFilter)
        {
            //variables
            TestCaseTemplateDisplayer methodDisplayer;
            IList<TestCaseTemplateDisplayer> methodDisplayerList = new List<TestCaseTemplateDisplayer>();
            //get Test Case Template List
            IList<TestCaseTemplate> testCaseTemplateList = TestCaseTemplateList.getInstance();
            //in case the class name filter is neither null nor empty, set a flag to use the class filter, otherwise set it to not
            bool useFilter = classNameFilter != null && !classNameFilter.Equals(String.Empty);
            //in case the list neither null or empty, execute the tests, otherwise do nothing.
            if (testCaseTemplateList != null && testCaseTemplateList.Count > 0)
            {
                //variables necessary for loop below
                Type type = null;
                MethodInfo[] methodList = null;
                //go through each Test Case Template List
                foreach (TestCaseTemplate testCaseTemplate in testCaseTemplateList)
                {
                    /*
                     * Use reflection to retrieve all methods. They should
                     * start with OtterUnitTestFramework.Test.TEST_METHOD_INITIAL_NAME. In
                     * case they do not, ignore them. 
                     * All selected methods (according to the criteria above) will be stored.
                     */

                    //assign value to type and get methods from it
                    type = testCaseTemplate.GetType();
                    methodList = type.GetMethods();
                    //go through all methods
                    if (methodList != null && methodList.Length > 0)
                    {
                        foreach (MethodInfo method in methodList)
                        {
                            //execute each method in case they fall in the criterea descripted above
                            if (method.Name.StartsWith(TestCaseTemplateConstants.TEST_METHOD_INITIAL_NAME) && method.GetParameters().Length == 0 && method.ReturnType.Equals(typeof(void)))
                            {
                                /*
                                 * In case the class filter is to be used do so, otherwise create
                                 * the Test Case Template Displayer and add it the the list
                                 */
                                if (useFilter)
                                {
                                    /*
                                     * The class filter is to be used therefore, see if the name of 
                                     * the filter class matches the name of the Test Case Template class.
                                     * If so, create the method info, otherwise do nothing.
                                     */
                                    if (classNameFilter.Equals(testCaseTemplate.GetType().Name))
                                    {
                                        //create a method info holder add it to the list
                                        methodDisplayer = new TestCaseTemplateDisplayer(testCaseTemplate, method);
                                        methodDisplayerList.Add(methodDisplayer); 
                                    }
                                }
                                else //no class filter is to be used therefore create the method info and add it to the list
                                {
                                    //create a method info holder add it to the list
                                    methodDisplayer = new TestCaseTemplateDisplayer(testCaseTemplate, method);
                                    methodDisplayerList.Add(methodDisplayer);
                                }
                            }
                        }
                    }
                }
            }
            //return list
            return methodDisplayerList;
        }

        /// <summary>
        /// This method is responsible for either checkin or unchecking
        /// all test within each Test Case Template Displayer shown Object.
        /// </summary>
        /// <param name="check">True in case all Test Case Displayers elements are to be checked, false otherwise</param>
        private void CheckOrUncheckAllMethodDisplays(bool check)
        {
            //set cursor
            this.Cursor = Cursors.WaitCursor;
            //get data source
            IList<TestCaseTemplateDisplayer> dataSource = (IList<TestCaseTemplateDisplayer>)_dgvTests.DataSource;
            //go through the data source list and (un)check all tests
            if (dataSource != null && dataSource.Count > 0)
            {
                for (int index = 0; index < dataSource.Count; index++)
                {
                    TestCaseTemplateDisplayer methodDiplayer = dataSource[index];
                    //(un)check method displayer ((un)check test)
                    methodDiplayer.ExecuteTest = check;
                    //update data source list
                    dataSource.RemoveAt(index);
                    dataSource.Insert(index, methodDiplayer);
                    //updata data grid view data source
                    _dgvTests.DataSource = dataSource;
                    //refresh grid
                    _dgvTests.Refresh();
                }
            }
            //reset cursor
            this.Cursor = Cursors.Default;
        }

        #endregion
    }
}