using System;
using System.Collections.Generic;
using System.Windows.Forms;
using OtterUnitTestFramework.Forms;
using OtterUnitTestFramework.Event;

namespace OtterUnitTestFramework.Public
{
    /// <summary>
    /// This class is responsible for launching the Framework. One has
    /// to implement it (implementing the method  public abstract void RegisterTests())
    /// , instantiate it and finally execute the method Execute Application.
    /// </summary>
    public class OtterUnitTestFrameworkApplication
    {
        /// <summary>
        /// Singleton instance
        /// </summary>
        private static OtterUnitTestFrameworkApplication _application;

        /// <summary>
        /// True in case the application is being executed, false otherwise
        /// </summary>
        private static bool _isApplicationBeingExecuted = false;

        /// <summary>
        /// Test Registration Instance
        /// </summary>
        private static TestRegistration _testRegistration = new TestRegistration();

        /// <summary>
        /// Getter and Setter for the Application Execution status. In case another
        /// instance of it is being executed, true will be returned, otherwise false
        /// will be given back.
        /// </summary>
        internal static bool IsApplicationBeingExecuted
        {
            get { return OtterUnitTestFrameworkApplication._isApplicationBeingExecuted; }
            set { OtterUnitTestFrameworkApplication._isApplicationBeingExecuted = value; }
        }

        /// <summary>
        /// Get a unique instance of OtterUnitTestFrameworkApplication (Singleton).
        /// </summary>
        /// <returns>Otter Unit Test Application Object Instance</returns>
        public static OtterUnitTestFrameworkApplication GetInstance()
        {
            //instantiate the object in case it is needed - singleton pattern
            if (_application == null)
            {
                _application = new OtterUnitTestFrameworkApplication();
            }
            return _application;
        }

        /// <summary>
        /// Private Constructor - avoid its instatiation becuase
        /// this is a singleton
        /// </summary>
        private OtterUnitTestFrameworkApplication()
        {
            //do nothing
        }

        /// <summary>
        /// This method executes the application. One should observe that only one instance of 
        /// this application is to be executed. thus, in case more than one instance is deployed, 
        /// an error message will be displayed and the application will not be executed.
        /// </summary>
        public void ExecuteApplication()
        {
            /*
             * In case there is no other instance of this application being 
             * run execute it, otherwise show an error message and do nothing.
             */
            if (!_isApplicationBeingExecuted)
            {
                //state one instance of the application is being run
                _isApplicationBeingExecuted = true;
                //execute application
                Application.EnableVisualStyles();
                Application.SetCompatibleTextRenderingDefault(false);
                Application.Run(new OtterUnitTestFrameworkForm());
            }
            else
            {
                //show error message
                MessageBox.Show(null, "There is another instance of the Otter Unit Test Application being run. Please close it before running another instance.", "Otter Unit Test Framework", MessageBoxButtons.OK, MessageBoxIcon.Stop);
            }
        }

        /// <summary>
        /// This method executes the application only when it is
        /// called from the Add-in.
        /// </summary>
        public void ExecuteApplicationInAddin()
        {
            /*
             * In case there is no other instance of this application being 
             * run execute it, otherwise show an error message and do nothing.
             */
            if (!_isApplicationBeingExecuted)
            {
                //state one instance of the application is being run
                _isApplicationBeingExecuted = true;
                //instantiate and show form
                OtterUnitTestFrameworkForm form = new OtterUnitTestFrameworkForm();
                form.Show();
            }
            else
            {
                //show error message
                MessageBox.Show(null, "There is another instance of the Otter Unit Test Application being run. Please close it before running another instance.", "Otter Unit Test Framework", MessageBoxButtons.OK, MessageBoxIcon.Stop);
            }
        }
 
        /// <summary>
        /// The main entry point for the application. It is only used
        /// when this framework is being developed.
        /// </summary>
        [STAThread]
        static void Main()
        {
            Application.EnableVisualStyles();
            Application.SetCompatibleTextRenderingDefault(false);
            Application.Run(new OtterUnitTestFrameworkForm());
        }

        /// <summary>
        /// Get the Test Registration Instance. With it, one can
        /// register and unregister Test Case Templates to be
        /// tested.
        /// </summary>
        public TestRegistration TestRegistration
        {
            get
            {
                return _testRegistration;
            }
        }
    }
}